package engine;

import java.util.Arrays;
import java.util.Comparator;

import board.Board;

public class MiniMax {

	private int maxDepth;
	private Board board;
	private double[] weights;
	private double minAlpha;
	private int maxAlphaI;

	public MiniMax(int maxDepth) {
		setMaxDepth(maxDepth);
	}
	
	/**
	 * @param maxDepth - maximun tree depth
	 */
	void setMaxDepth(int maxDepth) {
		if (maxDepth<=0)
			throw new IllegalArgumentException("maxDepth must be strictly positive");
		this.maxDepth = maxDepth;
	}
	
	/**
	 * Build the game tree and get the bounds (hence not accurate) of column scores, enabling us to retrieve accurately only the best score
	 * A value of null means the move is illegal
	 * @param board
	 * @param weights
	 * @return
	 * @see getMaxColumnScore
	 */
	public ColumnScore getMaxColumnScore(Board board, double[] weights) {
		return getMaxColumnScore(board, weights, true);
	}
	
	/**
	 * @param board
	 * @param weights
	 * @param useHeuristic - a boolean condition to order the moves not from 0 --> nCols, 
	 * but rather first those columns that the new checker will touch as many of its own pieces as possible
	 * @return ColumnScore - an instance of the best score  
	 */
	public ColumnScore getMaxColumnScore(Board board, double[] weights, boolean useHeuristic) {
		
		this.board = board;
		this.weights = weights;
		minAlpha = Double.NEGATIVE_INFINITY;
		double maxScore = minimax(0, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, true, useHeuristic);
		
		return new ColumnScore(maxAlphaI, maxScore);
		
	}

	/**
	 * Build the game tree and get the accurate column scores. This differs from getColumnScores in that it minimax's each child separately, so
	 * there is no inter-children pruning, which would cause some columns not to be fully analyzed - not calculating their accurate values
	 * A value of null means the move is illegal 
	 * @param board
	 * @param weights
	 * @return an array with scores for every move
	 */
	public Double[] getColumnScores(Board board, double[] weights) {

		this.board = board;
		this.weights = weights;

		Double[] scores = new Double[Constants.nCols];

		for(int i=0; i<Constants.nCols; ++i) {
			if (!board.isLegalMove(i)) {
				scores[i] = null;
			}
			else {
				board.setMove(i, false);
				scores[i] = minimax(1, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, false, true);
				board.popMove();
			}
		}

		return scores;
		
	}

	/**
	 * Recursive implementation where we step through (and build it along the way) the game tree, maintaining alpha-beta values and pruning
	 * @param depth
	 * @param alpha
	 * @param beta
	 * @param isMax
	 * @param heuristic - a boolean condition to order the moves not from 0 --> nCols, 
	 * but rather first those columns that the new checker will touch as many of its own pieces as possible
	 */
	private double minimax(int depth, double alpha, double beta, boolean isMax, boolean heuristic) {
		
		// if we've reached the bottom (i.e. leaves) of our tree either because of the search depth criterion, or because of the rules of the game
		// this is the terminating condition for the recursion
		if (depth >= maxDepth || board.winner() != -1)
			return Sarsa.calculateScore(weights, board.getFeatureVector());
		
		
		// IMPORTANT HEURISTIC
		// order the moves not from 0 --> nCols, but rather first those columns that the new checker will touch as many of its own pieces as possible
		
		ColumnScore columns[] = new ColumnScore[Constants.nCols];
		for(int j=0; j<Constants.nCols; ++j) {
			// could do null instead of -1, although that makes the comparator more complicated, and -1 is appropriate since countAdjacent is non-negative
			columns[j] = new ColumnScore(j, board.isLegalMove(j) ? (double)board.countAdjacent(j, isMax) : -1);
		}
		if (heuristic) Arrays.sort(columns, new Comparator<ColumnScore>() {
			@Override
			public int compare(ColumnScore o1, ColumnScore o2) {
				return o2.score.compareTo(o1.score); // descending!
			}
		});
		
		if (isMax) {
			for(int j=0; j<Constants.nCols; ++j) {
				if (columns[j].score == -1)
					continue;
				
				board.setMove(columns[j].col, false);
				
				// DFS in order to allow for alpha-beta pruning
				double child = minimax(depth+1, alpha, beta, false, heuristic);

				if (child > alpha) {
					alpha = child;
					// we are looking for the actual best move, which updates only on the first level
					if (depth == 0 && alpha > minAlpha) {
						minAlpha = alpha;
						maxAlphaI = columns[j].col;
					}
				}
				
				board.popMove();

				// PRUNING
				if (alpha >= beta) {
					//System.out.println("Prune: "+alpha+" >= "+beta);
					break;
				}
				
			}
			
			return alpha;
		}
		else {
			for(int j=0; j<Constants.nCols; ++j) {
				
				if (columns[j].score == -1)
					continue;
				
				board.setMove(columns[j].col, true);
				
				// DFS in order to allow for alpha-beta pruning
				double child = minimax(depth+1, alpha, beta, true, heuristic);

				if (child < beta)
					beta = child;
			
				board.popMove();

				// PRUNING
				if (alpha >= beta) {
					//System.out.println("Prune: "+alpha+" >= "+beta);
					break;
				}
				
			}
			
			return beta;
		}
		
	}
	
	/**
	 * A class holding a column index and a score for a move in the column
	 */
	static class ColumnScore {
		private int col;
		private Double score;
		public ColumnScore(int col, Double n) {
			this.col = col;
			this.score = n;
		}
		public int getCol() {
			return col;
		}
		public Double getScore() {
			return score;
		}
	}
	
}
