package runtime;

import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.ProgressBar;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Tree;
import org.eclipse.swt.widgets.TreeColumn;
import org.eclipse.swt.widgets.TreeItem;

import runtime.Manager.Results;
import runtime.Manager.TrainingData;
import engine.Constants;
import engine.Sarsa;

public class ManagerGUI {

	private static ExecutorService thread = Executors.newCachedThreadPool();

	private static Display display;
	private static Shell shell;
	private static Tree convergence;
	private static Label resultsLabel;
	private static ProgressBar trainProgress, testProgress;
	private static Button trainStart, trainStop, trainSave, trainAll, trainStopAll, test, testAll;
	private static boolean isTraining, isTrainingAll;
	static String trainingBoard;

	private static Sarsa[] learners;
	
	static void buildGUI() {

		display = new Display();
		shell = new Shell(display);

		shell.setText("Connect Four Manager");
		shell.setImage(new Image(display, "resources\\icon.png"));
		shell.setBackground(display.getSystemColor(SWT.COLOR_WHITE));
		shell.setBackgroundMode(SWT.INHERIT_FORCE);
		shell.setSize(700, 600);

		// center on screen
		Rectangle screen = display.getPrimaryMonitor().getBounds(), window = shell.getBounds();
		shell.setLocation(screen.x + (screen.width - window.width)/2, screen.y + (screen.height - window.height)/2);

		shell.setLayout(new GridLayout(1,true));
		
		Composite trainArea = new Composite(shell, SWT.NONE);
		trainArea.setLayout(new GridLayout(5,false));
    	trainArea.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
		
		final Combo trustFunction = new Combo(trainArea, SWT.DROP_DOWN | SWT.READ_ONLY);
		trustFunction.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
		trustFunction.add("Trust_0.5");
		trustFunction.add("Trust_1");
		trustFunction.add("Trust_2");
		trustFunction.select(0);
		
		final Combo features = new Combo(trainArea, SWT.DROP_DOWN | SWT.READ_ONLY);
    	features.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
    	
		for(String board : Manager.boards.values())
			features.add(board);
		features.select(0);
		
		trainStart = new Button(trainArea, SWT.NONE);
		trainStart.setText("Train");
		
		trainStop = new Button(trainArea, SWT.NONE);
		trainStop.setText("Stop");
		trainStop.setEnabled(false);
		
		trainSave = new Button(trainArea, SWT.NONE);
		trainSave.setText("Save");
		trainSave.setEnabled(false);

		Composite testAllArea = new Composite(shell, SWT.NONE);
    	testAllArea.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
		testAllArea.setLayout(new GridLayout(2,true));

		trainAll = new Button(testAllArea, SWT.NONE);
		trainAll.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
		trainAll.setText("Train All");
		
		trainStopAll = new Button(testAllArea, SWT.NONE);
		trainStopAll.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
		trainStopAll.setText("Stop");
		trainStopAll.setEnabled(false);
		
		trainProgress = new ProgressBar(shell, SWT.SMOOTH | SWT.HORIZONTAL);
		trainProgress.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
		trainProgress.setMinimum(0);
		
		convergence = new Tree(shell, SWT.BORDER | SWT.FULL_SELECTION);
		convergence.setLinesVisible(true);
		convergence.setHeaderVisible(true);
    	convergence.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));

    	String[] columns = {"Player","Iterations","Time","Weights"};

    	for (String col : columns) {
    		new TreeColumn(convergence, SWT.NONE).setText(col);
    	}
    	packCols();
		
		trainStart.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				trainStart.setEnabled(false);
				trainStop.setEnabled(true);
				trainSave.setEnabled(false);
				trainAll.setEnabled(false);
				trainStopAll.setEnabled(false);

				final int trust = trustFunction.getSelectionIndex();
				
				isTraining = true;
				trainingBoard = getBoardIndex(features.getText());
				convergence.removeAll();
				trainProgress.setSelection(0);
				trainProgress.setMaximum(Constants.convergedT*2);
				
				thread.execute(new Runnable() {
					@Override
					public void run() {
						Manager.train(trainingBoard, trust);
					}
				});
			}
		});
		trainStop.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				stop();
			}
		});
		trainSave.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				trainStop.setEnabled(false);
				trainStart.setEnabled(true);
				trainSave.setEnabled(false);

				final int trust = trustFunction.getSelectionIndex();
				
				thread.execute(new Runnable() {
					@Override
					public void run() {
						Manager.save(learners[0].getWeights(), learners[1].getWeights(), learners[0].getIterations(), trainingBoard, trust);
					}
				});
			}
		});

		trainAll.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				trainStop.setEnabled(false);
				trainStart.setEnabled(false);
				trainSave.setEnabled(false);
				trainAll.setEnabled(false);
				trainStopAll.setEnabled(true);
				isTraining = isTrainingAll = true;

				trainProgress.setSelection(0);
				trainProgress.setMaximum(Constants.convergedT*features.getItemCount()*trustFunction.getItemCount());

				for(int t=0; t<trustFunction.getItemCount();++t) {
					final int trust = t;
					for(String board : features.getItems()) {
					
						final String b = board;
						
						thread.execute(new Runnable() {
							@Override
							public void run() {
								Manager.train(getBoardIndex(b), trust);
							}
						});
					
					}
				}
				
			}
		});
		trainStopAll.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				stopTrainAll();
			}
		});
		
		Composite testArea = new Composite(shell, SWT.NONE);
    	testArea.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
		testArea.setLayout(new GridLayout(2,true));

		final Combo opponent1 = new Combo(testArea, SWT.DROP_DOWN | SWT.READ_ONLY);
		final Combo opponent2 = new Combo(testArea, SWT.DROP_DOWN | SWT.READ_ONLY);
		opponent1.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
    	opponent2.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));

		opponent1.add("Random");
		opponent2.add("Random");
		opponent1.add("Manual");
		opponent2.add("Manual");
		for(String board : Manager.boards.values()) {
			opponent1.add(board);
			opponent2.add(board);
		}
		opponent1.select(0);
		opponent2.select(0);

		final Combo opponent1Weights = new Combo(testArea, SWT.DROP_DOWN | SWT.READ_ONLY);
		final Combo opponent2Weights = new Combo(testArea, SWT.DROP_DOWN | SWT.READ_ONLY);
		opponent1Weights.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
		opponent2Weights.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
		
		opponent1.addSelectionListener(listenerSavedWeights(opponent1, opponent1Weights));
		opponent2.addSelectionListener(listenerSavedWeights(opponent2, opponent2Weights));
		
		test = new Button(shell, SWT.NONE);
		test.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
		test.setText("Play/Test Opponents");
		test.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				test.setEnabled(false);
				testAll.setEnabled(false);
				final String player1 = getPlayerIndex(opponent1.getText()), player2 = getPlayerIndex(opponent2.getText());
				if (isValidSelection(player1, opponent1Weights.getText()) && isValidSelection(player2, opponent2Weights.getText())) {	
					final double[][] weights1 = getWeights(player1, opponent1Weights.getText()), weights2 = getWeights(player2, opponent2Weights.getText());
					
					testProgress.setMaximum(Manager.getNumGames(player1, player2));
					testProgress.setSelection(0);
					resultsLabel.setText("");
					
					thread.execute(new Runnable() {
						@Override
						public void run() {
							Manager.test(player1, player2, weights1, weights2);
						}
					});
				}
			}
		});
		
		testAll = new Button(shell, SWT.NONE);
		testAll.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
		testAll.setText("Test All Saved Weights");
		testAll.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				test.setEnabled(false);
				testAll.setEnabled(false);
				thread.execute(new Runnable() {
					@Override
					public void run() {
						Manager.testAll();
					}
				});
			}
		});
		
		testProgress = new ProgressBar(shell, SWT.SMOOTH | SWT.HORIZONTAL);
		testProgress.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, false));
		testProgress.setMinimum(0);
		
		resultsLabel = new Label(shell, SWT.NONE);
		resultsLabel.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true));
		
		Manager.loadTraining();

		shell.open();

		while (!shell.isDisposed()) {
			if (!display.readAndDispatch())
				display.sleep();
		}

		thread.shutdownNow();
	}

	protected static void stopTrainAll() {
		isTraining = false;
		try { // hack to avoid a whole bunch of variables and conditions. works nicely :)
			Thread.sleep(100);
		} catch (InterruptedException e1) {
			e1.printStackTrace();
		}
		stop(false);
	}

	protected static void stop() {
		stop(true);
	}
	protected static void stop(final boolean b) {
		display.asyncExec(new Runnable() {
			@Override
			public void run() {
				trainStop.setEnabled(false);
				trainStart.setEnabled(true);
				trainSave.setEnabled(b);
				trainAll.setEnabled(true);
				trainStopAll.setEnabled(false);
			}
		});
		
		isTraining = isTrainingAll = false;
	}
	
	private static Runnable bumpTrainAllProgress = new Runnable() {
		@Override
		synchronized public void run() {
			trainProgress.setSelection(trainProgress.getSelection()+Constants.trainAllSaveInterval);
			if (trainProgress.getMaximum()-trainProgress.getSelection() < Constants.trainAllSaveInterval) {
				trainProgress.setSelection(trainProgress.getMaximum());
				stopTrainAll();
			}
		}
	};

	public static void finishedTraining(Sarsa[] learners) {
		ManagerGUI.learners = learners;
		stop();
	}
	
	public static void saveInterval(final TrainingData data, final String board) {
		thread.execute(new Runnable() {
			@Override
			public void run() {
				Manager.save(data.weights[0], data.weights[1], data.t, board, data.trust);
			}
		});
		display.asyncExec(bumpTrainAllProgress);
	}
	
	public static boolean isTrainingAll() {
		return isTrainingAll;
	}
	
	static boolean isValidSelection(String player, String weightsDesription) {
		if (!(player.equals("Random") || player.equals("Manual")) && weightsDesription.isEmpty()) {
			MessageBox dialog = new MessageBox(shell);
			dialog.setText("Select Trained Weights");
			dialog.setMessage("For Non Manual/Random Players, you need to select valid weights. If none are offered, then train the configuration above.");
			dialog.open();
			return false;
		}
		return true;
	}

	private static SelectionListener listenerSavedWeights(final Combo combo, final Combo comboWeights) {
		return new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				comboWeights.removeAll();
				if (!combo.getText().equals("Random") && !combo.getText().equals("Manual")) {
					for(TrainingData data : Manager.trainingData.get(getBoardIndex(combo.getText()))) {
						comboWeights.add(getWeightsDescription(data));
					}
					comboWeights.select(0);
				}
			}
		};
	}

	private static SimpleDateFormat dateFormatter = new SimpleDateFormat("d/M HH:mm:ss");
	protected static String getWeightsDescription(TrainingData data) {
		return dateFormatter.format(data.date) + " trust:"+ data.trust + " iterations:" + data.t;
	}

	protected static String getBoardIndex(String description) {
		for(Map.Entry<String,String> entry : Manager.boards.entrySet()) {
			if (entry.getValue().equals(description)) {
				return entry.getKey();
			}
		}
		return null;
	}
	
	protected static String getPlayerIndex(String player) {
		return !player.equals("Manual") && !player.equals("Random") ? getBoardIndex(player) : player;
	}
	
	protected static double[][] getWeights(String player, String weightsDescription) {
		if (!player.equals("Manual") && !player.equals("Random")) {
			for(TrainingData data : Manager.trainingData.get(player)) {
				if (getWeightsDescription(data).equals(weightsDescription)) {
					return data.weights;
				}
			}
		}
		return null;
	}

	static boolean isTraining() {
		return !shell.isDisposed() && isTraining;
	}
	
	private static void packCols() {
    	for (int i=0; i<convergence.getColumnCount(); i++) {
    		convergence.getColumn(i).pack();
    	}
	}

	public static boolean isActive() {
		return !shell.isDisposed();
	}
	
	//private static double[] oldWeights;
	public static void displayConvergenceValues(final int t, double[] weights, final long time, final boolean didIStart) {
		
		// calculate delta
		/*if(didIStart) {
			double delta = 0;
			if (oldWeights != null) {
				for(int i=0; i<weights.length; ++i) {
					double d = weights[i] - oldWeights[i];
					delta += d*d;
				}
				System.out.println(t/100000.0 + " " +delta);
			}
			oldWeights = weights.clone();
		}*/
		
		DecimalFormat decimalFormat = new DecimalFormat("0.#####");
		final StringBuffer s = new StringBuffer();
		for(int i=0; i<weights.length; ++i) {
			s.append(decimalFormat.format(weights[i]));
			s.append(",");
		}
		s.deleteCharAt(s.length()-1);
		
		display.asyncExec(new Runnable() {
			@Override
			public void run() {

				if (!isTrainingAll()) trainProgress.setSelection(trainProgress.getSelection()+Constants.outputInterval);
				
				TreeItem update = new TreeItem(convergence, SWT.NONE);
				update.setText(new String[]{didIStart?"1":"2",Integer.toString(t),Long.toString(time),s.toString()});

				packCols();
			}
		});
		
	}

	public static void displayResults(final String s) {
		display.asyncExec(new Runnable() {
			@Override
			public void run() {
				resultsLabel.setText(s);
				test.setEnabled(true);
				testAll.setEnabled(true);
			}
		});
	}

	synchronized public static void incTest() {
		display.asyncExec(new Runnable() {
			@Override
			public void run() {
				testProgress.setSelection(testProgress.getSelection()+1);
				if (testProgress.getSelection()==testProgress.getMaximum() && !test.getEnabled() && Manager.thread!=null && !Manager.thread.isShutdown()) {
					test.setEnabled(true);
					testAll.setEnabled(true);

					Manager.thread.shutdownNow();

					printResults(Manager.results); // to compare, just to be sure that the sorting didn't mess up the results
					System.out.println();
					System.out.println();
					
					sortResults(Manager.results,1);
					transpose(Manager.results);
					sortResults(Manager.results,-1);
					transpose(Manager.results);
					
					printResults(Manager.results);
				}
			}
		});
	}
	
	private static void transpose(Results[][] results) {
		for(int i=0;i<results.length;++i) {
			for(int j=i+1;j<results.length;++j) {
				Results temp = results[i][j];
				results[i][j] = results[j][i];
				results[j][i] = temp;
			}
		}
	}

	private static void printResults(Results[][] results) {
		for(int i=0;i<results.length;++i) {
			System.out.print(results[i][0].p1.board);
			System.out.print(results[i][0].p1.name);
			System.out.print("&");
		}
		System.out.println();
		System.out.println();
		for(int i=0;i<results.length;++i) {
			System.out.print(results[0][i].p2.board);
			System.out.print(results[0][i].p2.name);
			System.out.print("&");
		}
		System.out.println();
		System.out.println();
		System.out.println("Wins:");
		for(int i=0;i<results.length;++i) {
			for(int j=0;j<results.length;++j) {
				System.out.print(results[i][j].wins);
				System.out.print("&");
			}
			System.out.println();
		}
		System.out.println();
		System.out.println("Ties:");
		for(int i=0;i<results.length;++i) {
			for(int j=0;j<results.length;++j) {
				System.out.print(results[i][j].ties);
				System.out.print("&");
			}
			System.out.println();
		}
	}

	private static void sortResults(Results[][] results, final int m) {
		Arrays.sort(results, new Comparator<Results[]>() {
			@Override
			public int compare(Results[] o1, Results[] o2) {
				int n1=0, n2=0;
				for(Results r : o1) {
					n1 += r.wins;
					n1 += r.ties;
				}
				for(Results r : o2) {
					n2 += r.wins;
					n2 += r.ties;
				}
				return (n2-n1)*m;
			}
		});
	}



	public static void setTestAllN(final int n) {
		display.asyncExec(new Runnable() {
			@Override
			public void run() {
				testProgress.setMaximum(n*n*Constants.numTestsForWinningStatistics);
				testProgress.setSelection(0);
			}
		});
	}
	
}
